import JsxSyntax from '@babel/plugin-syntax-jsx';
import { addNamed } from '@babel/helper-module-imports';
import { identifier, isJSXNamespacedName, isJSXMemberExpression, callExpression, stringLiteral, binaryExpression, objectProperty, isJSXSpreadAttribute, spreadElement, objectExpression, isJSXExpressionContainer, isJSXEmptyExpression, nullLiteral, booleanLiteral, isJSXText, isJSXElement, isJSXFragment, isJSXSpreadChild, isSpreadElement, arrayExpression } from '@babel/types';

const JSX_LIBRARY_NAME = '@whatsup/jsx';
const FRAGMENT_COMPONENT_NAME = 'Fragment';
const JSX_LIBRARY_FACTORIES = {
  HTML: 'html',
  SVG: 'svg',
  Component: 'component'
};
const IS_SVG_REGEX = /^(?:sv|an|c(?:ir|l)|de[fs]|el|f(?:e|il|ore)|g$|ima|line|m(?:a(?:rke|sk)|etad|p)|p(?:at[ht]|oly[gl])|r[ae]|s(?:to|[wy])|t(?:ext(?:$|P)|s)|us|vie)/;
const IS_TESTING = process.env.PLUGIN_TESTING === 'on';

/// <reference path="./@types/index.d.ts" />
const VOID = identifier('undefined');
function index () {
  return {
    inherits: JsxSyntax,
    visitor: {
      JSXElement(path) {
        const {
          node
        } = path;
        const {
          openingElement,
          children
        } = node;
        const {
          name,
          attributes
        } = openingElement;

        if (isJSXNamespacedName(name)) {
          throw new Error('Namespaced openingElement name is not supported');
        }

        if (isJSXMemberExpression(name)) {
          throw new Error('OpeningElement name as member expression is not supported');
        } // Node from @babel/core & from @babel/types incompatible


        replaceJSXAstToMutatorFactoryCallExpression(path, name.name, attributes, children);
      },

      JSXFragment(path) {
        const {
          node
        } = path;
        const {
          children
        } = node; // Node from @babel/core & from @babel/types incompatible

        replaceJSXAstToMutatorFactoryCallExpression(path, FRAGMENT_COMPONENT_NAME, [], children);
      }

    }
  };
}

function replaceJSXAstToMutatorFactoryCallExpression(path, name, attributes, children) {
  const factory = getFactory(name);
  const callee = createCalleeImport(path, factory);
  const type = createType(path, name);
  const {
    key,
    props,
    ref,
    onMount,
    onUnmount
  } = parseAttributes(attributes, parseChildren(children));
  const args = createCalleeArgs(type, key, props, ref, onMount, onUnmount);
  const expression = callExpression(callee, args); // Node from @babel/core & from @babel/types incompatible

  path.replaceWith(expression);
}

function createCalleeImport(path, factory) {
  return createImport(path, JSX_LIBRARY_NAME, factory);
}

function createCalleeArgs(type, key, props, ref, onMount, onUnmount) {
  const salt = createSalt();
  const args = [type, key ? binaryExpression('+', salt, key) : salt, props, ref, onMount, onUnmount];
  let popping = true;

  for (let i = args.length; i > 0; i--) {
    if (args[i - 1]) {
      popping = false;
    } else if (popping) {
      args.pop();
    } else {
      args[i - 1] = VOID;
    }
  }

  return args;
}

function getFactory(name) {
  const {
    HTML,
    SVG,
    Component
  } = JSX_LIBRARY_FACTORIES;
  return isComponent(name) ? Component : isSVG(name) ? SVG : HTML;
}

function createFragmentImport(path) {
  return createImport(path, JSX_LIBRARY_NAME, FRAGMENT_COMPONENT_NAME);
}

function createImport(path, importSource, method) {
  return addNamed(path, method, importSource, {
    importedType: 'es6'
  });
}

function createType(path, name) {
  if (isFragment(name)) {
    return createFragmentImport(path);
  }

  if (isComponent(name)) {
    return identifier(name);
  }

  return stringLiteral(name);
}

function createSalt() {
  return stringLiteral(generateUid() + '_');
}

function parseAttributes(attributes, children) {
  let key = undefined;
  let ref = undefined;
  let onMount = undefined;
  let onUnmount = undefined;
  const members = [];

  if (children) {
    const member = objectProperty(identifier('children'), children);
    members.push(member);
  }

  for (const attr of attributes) {
    if (isJSXSpreadAttribute(attr)) {
      const member = spreadElement(attr.argument);
      members.push(member);
      continue;
    }

    const name = parseAttrName(attr.name);
    const value = parseAttrValue(attr.value);

    if (value !== undefined) {
      if (name === 'key') {
        key = value;
      } else if (name === 'ref') {
        ref = value;
      } else if (name === 'onMount') {
        onMount = value;
      } else if (name === 'onUnmount') {
        onUnmount = value;
      } else {
        const prop = isJSXNamespacedName(attr.name) ? stringLiteral(name) : identifier(name);
        const member = objectProperty(prop, value);
        members.push(member);
      }
    }
  }

  const props = members.length ? objectExpression(members) : undefined;
  return {
    key,
    props,
    ref,
    onMount,
    onUnmount
  };
}

function parseAttrName(name) {
  if (isJSXNamespacedName(name)) {
    return name.namespace.name + ':' + name.name.name;
  }

  return name.name;
}

function parseAttrValue(value) {
  if (isJSXExpressionContainer(value)) {
    if (isJSXEmptyExpression(value.expression)) {
      // Uncoveraged. JSX attributes must only be assigned a non-empty expression
      return nullLiteral();
    }

    return value.expression;
  }

  if (value === null) {
    return booleanLiteral(true);
  }

  return value;
}

function parseChildren(children) {
  const members = [];
  const {
    length
  } = children;

  for (let i = 0; i < length; i++) {
    const child = children[i];

    if (isJSXText(child)) {
      let value = child.value.replace(/ +/g, ' ').replace(/ ?\n ?/g, '\n');

      if (i === 0) {
        value = value.trimStart();
      }

      if (i === length - 1) {
        value = value.trimEnd();
      }

      if (value !== '' && value !== '\n') {
        members.push(stringLiteral(value.replace(/\n+/g, ' ')));
      }

      continue;
    }

    if (isJSXElement(child) || isJSXFragment(child)) {
      members.push(child);
      continue;
    }

    if (isJSXExpressionContainer(child) && !isJSXEmptyExpression(child.expression)) {
      members.push(child.expression);
      continue;
    }

    if (isJSXSpreadChild(child)) {
      members.push(spreadElement(child.expression));
      continue;
    }
  }

  if (!members.length) {
    return undefined;
  }

  if (members.length === 1 && !isSpreadElement(members[0])) {
    return members[0];
  }

  return arrayExpression(members);
}

function generateUid() {
  return IS_TESTING ? 'UniqueId' : Math.random().toString(36).slice(-5);
}

function isComponent(name) {
  const firstLetter = name.charAt(0);
  return firstLetter.toUpperCase() === firstLetter;
}

function isFragment(name) {
  return name === FRAGMENT_COMPONENT_NAME;
}

function isSVG(name) {
  return IS_SVG_REGEX.test(name);
}

export { index as default };
//# sourceMappingURL=index.esm.js.map
